from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Optional, Protocol, Tuple

import numpy as np
import torch
from torch import Tensor
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader

from ONTraC.utils.decorators import selective_args_decorator

from ..log import debug
from ..utils import round_epoch_filter
from ..utils.decorators import selective_args_decorator
from .loss_funs import masked_variance, within_cluster_variance_loss


class BatchTrain(ABC):
    """docstring for BatchTrain"""

    def __init__(self, model: torch.nn.Module, device: torch.device, data_loader: DataLoader) -> None:
        super(BatchTrain, self).__init__()
        self.model: torch.nn.Module = model
        self.device: torch.device = device
        self.data_loader: DataLoader = data_loader
        self.model = self.model.to(device=self.device)

    def __str__(self):
        return f"{self.__class__.__name__}(model='{self.model}', device='{self.device}', data_loader='{self.data_loader}')"

    def __repr__(self):
        return self.__str__()

    def train(self,
              max_epochs: int = 100,
              max_patience: int = 50,
              min_delta: float = 0,
              min_epochs: int = 100,
              *args,
              **kwargs) -> None:

        self.set_train_args(*args, **kwargs)

        self.model.train()
        min_loss = np.inf
        patience = 0
        best_params = self.model.state_dict()

        for epoch in range(max_epochs):
            train_loss = self.train_epoch(epoch=epoch)
            if np.isnan(train_loss):  # unexpected situation
                best_params = self.model.state_dict()
                break
            elif max_patience == 0:  # no early stopping
                best_params = self.model.state_dict()
            elif min_loss - train_loss < min_loss * min_delta:  # no improvement
                patience += 1
            else:  # improvement
                min_loss = train_loss
                patience = 0
                best_params = self.model.state_dict()
            # max_patience == 0 means no early stopping
            if max_patience != 0 and patience >= max_patience and epoch >= min_epochs:
                break
            if round_epoch_filter(epoch) and 'output' in kwargs:
                output_dir = kwargs['output']
                self.save(f'{output_dir}/epoch_{epoch + 1}.pt')
        self.model.load_state_dict(best_params)

    @abstractmethod
    def set_train_args(self) -> None:
        """Method that should be implemented by all derived classes."""
        raise NotImplementedError("The set_train_args method should be implemented by subclasses.")

    @abstractmethod
    def train_epoch(self, epoch: int) -> float:
        """Method that should be implemented by all derived classes."""
        raise NotImplementedError("The train_epoch method should be implemented by subclasses.")

    @abstractmethod
    def evaluate(self) -> Dict[str, np.floating]:
        """Method that should be implemented by all derived classes."""
        raise NotImplementedError("The evaluate method should be implemented by subclasses.")

    def predict(self, data: Data) -> Tuple[Tensor, ...] | Tensor:
        self.model.eval()
        with torch.no_grad():
            res = self.model.predict(data.x, data.adj, data.mask)  # type: ignore
        return res

    @abstractmethod
    def predict_dict(self, data: Data) -> Dict[str, Tensor]:
        """Method that should be implemented by all derived classes."""
        raise NotImplementedError("The predict_dict method should be implemented by subclasses.")

    def save(self, path: str) -> None:
        torch.save(self.model.state_dict(), path)

    def load(self, path: str) -> None:
        self.model.load_state_dict(torch.load(path))


class SubBatchTrainProtocol(Protocol):

    def train(self, *args, **kwargs) -> None:
        ...

    def evaluate(self) -> Dict[str, np.floating]:
        ...

    def predict(self, data: Data) -> Tuple[Tensor, ...] | Tensor:
        ...

    def predict_dict(self, data: Data) -> Dict[str, Tensor]:
        ...

    def save(self, path: str) -> None:
        ...

    def load(self, path: str) -> None:
        ...


class GPBatchTrain(BatchTrain):
    """
    Graph Pooling batch training class.
    """

    @selective_args_decorator
    def set_train_args(self,
                       optimizer: torch.optim.Optimizer,
                       modularity_loss_weight: float = 1,
                       purity_loss_weight: float = 0,
                       regularization_loss_weight: float = 1,
                       ortho_loss_weight: float = 0,
                       inspect_funcs: Optional[List[Callable]] = None) -> None:
        self.optimizer = optimizer
        self.spectral_loss_weight = modularity_loss_weight
        self.ortho_loss_weight = ortho_loss_weight
        self.cluster_loss_weight = regularization_loss_weight
        self.feat_similarity_loss_weight = purity_loss_weight
        self.inspect_funcs = inspect_funcs

    def cal_loss(self, spectral_loss, ortho_loss, cluster_loss, data, s) -> Tuple[Tensor, ...]:
        spectral_loss = self.spectral_loss_weight * spectral_loss
        ortho_loss = self.ortho_loss_weight * ortho_loss * np.sqrt(2)
        cluster_loss = self.cluster_loss_weight * cluster_loss / (np.sqrt(self.model.k) - 1)
        feat_similarity_loss = within_cluster_variance_loss(x=data.x, s=s, mask=data.mask)
        total_var = masked_variance(x=data.x, mask=data.mask)
        feat_similarity_loss = self.feat_similarity_loss_weight * feat_similarity_loss
        loss = spectral_loss + ortho_loss + cluster_loss + feat_similarity_loss

        return loss, spectral_loss, ortho_loss, cluster_loss, feat_similarity_loss

    def train_epoch(self, epoch: int) -> float:
        self.model.train()
        train_loss = 0
        for batch, data in enumerate(self.data_loader):
            # debug(f'epoch {epoch+1}, batch {batch+1} start.')
            data = data.to(self.device)
            s, out, out_adj, spectral_loss, ortho_loss, cluster_loss = self.model(data.x, data.adj, data.mask)
            loss, spectral_loss, ortho_loss, cluster_loss, feat_similarity_loss = self.cal_loss(
                spectral_loss, ortho_loss, cluster_loss, data, s)
            loss.backward()
            self.optimizer.step()
            self.optimizer.zero_grad()
            # debug(f'epoch {epoch+1}, batch {batch+1} end.')
            if self.inspect_funcs is not None:
                for inspect_func in self.inspect_funcs:
                    inspect_func(
                        epoch=epoch + 1,
                        batch=batch + 1,
                        data=data,
                        s=s,
                        out=out,
                        out_adj=out_adj,
                        loss=loss,
                        spectral_loss=spectral_loss,
                        # ortho_loss=ortho_loss,
                        cluster_loss=cluster_loss,
                        feat_similarity_loss=feat_similarity_loss)
            train_loss += loss.item()
        return train_loss

    def evaluate(self) -> Dict[str, np.floating]:
        """
        Evaluate the model.
        :return: results_dict
        """
        spectral_loss_list, ortho_loss_list, cluster_loss_list = [], [], []
        # bin_spectral_loss_list, bin_ortho_loss_list, bin_cluster_loss_list = [], [], []
        feat_similarity_loss_list = []
        loss_list = []
        self.model.eval()
        with torch.no_grad():
            for data in self.data_loader:
                data = data.to(self.device)
                s, out, out_adj, spectral_loss, ortho_loss, cluster_loss = self.model.evaluate(
                    data.x, data.adj, data.mask)
                loss, spectral_loss, ortho_loss, cluster_loss, feat_similarity_loss = self.cal_loss(
                    spectral_loss, ortho_loss, cluster_loss, data, s)

                spectral_loss_list.append(spectral_loss.item())
                # ortho_loss_list.append(ortho_loss.item())
                cluster_loss_list.append(cluster_loss.item())
                feat_similarity_loss_list.append(feat_similarity_loss.item())
                loss_list.append(loss.item())
        spectral_loss = np.mean(spectral_loss_list)
        # ortho_loss = np.mean(ortho_loss_list)
        cluster_loss = np.mean(cluster_loss_list)
        feat_similarity_loss = np.mean(feat_similarity_loss_list)
        loss = np.mean(loss_list)
        results_dict = {
            'modularity_loss': spectral_loss,
            'purity_loss': feat_similarity_loss,
            'regularization_loss': cluster_loss,
            # 'ortho_loss': ortho_loss,
            'total_loss': loss
        }
        return results_dict

    def predict_dict(self, data: Data) -> Dict[str, Tensor]:
        self.model.eval()
        with torch.no_grad():
            s, out, out_adj = self.model.predict(data.x, data.adj, data.mask)
            z = self.model.predict_embed(data.x, data.adj, data.mask)
        return {'z': z, 's': s, 'out': out, 'out_adj': out_adj}

    def predict_embed(self, data: Data) -> Tensor:
        self.model.eval()
        with torch.no_grad():
            z = self.model.predict_embed(data.x, data.adj, data.mask)  # type: ignore
        return z


__all__ = ['SubBatchTrainProtocol', 'GPBatchTrain']



import inspect
import os
from functools import wraps
from typing import Callable

from ..log import debug


def get_default_args(func) -> dict:
    """Get default arguments of a function.

    Args:
        func (function): A function.

    Returns:
        dict: Default arguments of the function.
    """
    return {
        k: None if v.default is inspect.Parameter.empty else v.default
        for k, v in inspect.signature(func).parameters.items()
    }


def selective_args_decorator(func) -> Callable:
    """Decorator that allows a function to accept only a subset of its arguments.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # debug(f'selective_args_decorator: args: {args}')
        # debug(f'selective_args_decorator: kwargs: {list(kwargs.keys())}')
        default_args = {}
        for key, value in inspect.signature(func).parameters.items():
            # debug(f'key: {key}, value: {value}')
            if value.default is inspect.Parameter.empty:  # without default value
                if key in ['args', 'kwargs', 'self']:  # skip
                    continue
                elif key in kwargs:  # with input value
                    default_args[key] = kwargs[key]
                else:  # without input value
                    raise TypeError(f'{func.__name__}() missing 1 required positional argument: {key}')
            else:  # with default value
                default_args[key] = kwargs[key] if key in kwargs else value.default
        if 'kwargs' in default_args:
            default_args.update(kwargs)
        # debug(f'default_args: {list(default_args.keys())}')

        # call the function with the updated arguments
        return func(*args, **default_args)

    return wrapper


def epoch_filter_decorator(func: Callable) -> Callable:
    """
    Epoch filter decorator
    :param func: function
    :return: function
    """

    @wraps(func)
    def wrapper(*args, epoch_filter: Callable, **kwargs) -> None:  # add epoch_filter argument
        # debug(f'epoch_filter_decorator: args: {args}')
        # debug(f'epoch_filter_decorator: kwargs: {list(kwargs.keys())}')
        epoch: int = kwargs.get('epoch')  # type: ignore
        output_dir: str = kwargs.get('output_dir')  # type: ignore

        # Check the epoch_filter condition
        if epoch_filter(epoch):
            # Check if output_dir of this epoch exists
            epoch_output_dir: str = os.path.join(output_dir, f'Epoch_{epoch}')
            if not os.path.exists(epoch_output_dir):
                os.mkdir(epoch_output_dir)
            # Call the original function if condition is True
            return func(*args, **kwargs)
        else:
            # Otherwise, return None or some default value
            return None

    # Update the function signature
    sig = inspect.signature(func)
    parms = list(sig.parameters.values())
    parms.append(
        inspect.Parameter(name='epoch_filter',
                          kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          default=inspect.Parameter.empty,
                          annotation=Callable))
    wrapper.__signature__ = sig.replace(parameters=parms)  # type: ignore

    return wrapper


import os
import sys
from optparse import OptionGroup, OptionParser, Values
from typing import List, Optional

from ..log import *


def add_IO_options_group(optparser: OptionParser, io_options: Optional[List[str]]) -> None:
    """
    Add I/O options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """
    if io_options is None:
        return
    # I/O options group
    group_io = OptionGroup(optparser, "IO")
    if 'dataset' in io_options:
        group_io.add_option('-d', '--dataset', dest='dataset', type='string', help='Original input dataset.')
    if 'preprocessing_dir' in io_options:
        group_io.add_option('--preprocessing-dir',
                            dest='preprocessing_dir',
                            type='string',
                            help='Directory for preprocessing outputs.')
    if 'GNN_dir' in io_options:
        group_io.add_option('--GNN-dir', dest='GNN_dir', type='string', help='Directory for the GNN output.')
    if 'NTScore_dir' in io_options:
        group_io.add_option('--NTScore-dir', dest='NTScore_dir', type='string', help='Directory for the NTScore output')

    optparser.add_option_group(group_io)


def validate_io_options(optparser: OptionParser, options: Values, io_options: Optional[List[str]]) -> None:
    """Validate IO options from a OptParser object.

    Ret: None
    """
    if io_options is None:
        return
    if 'dataset' in io_options:
        if not options.dataset:
            error('Please provide a dataset.')
            optparser.print_help()
            sys.exit(1)
        if not os.path.isfile(options.dataset):
            error(f'The input file ({options.dataset}) you given does not exist.')
            optparser.print_help()
            sys.exit(1)
        if not options.dataset.endswith(('csv', 'csv.gz')):
            error(f'The input file ({options.dataset}) should be in csv format.')
            optparser.print_help()
            sys.exit(1)

    if 'preprocessing_dir' in io_options:
        if not options.preprocessing_dir:
            error('Please provide a directory for preprocessing outputs.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.preprocessing_dir):
            warning(f'The directory ({options.preprocessing_dir}) you given already exists. It will be overwritten.')
        else:
            info(f'Creating directory: {options.preprocessing_dir}')
            os.makedirs(options.preprocessing_dir)

    if 'GNN_dir' in io_options:
        if not options.GNN_dir:
            error('Please provide a directory for the GNN output.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.GNN_dir):
            warning(f'The directory ({options.GNN_dir}) you given already exists. It will be overwritten.')
        else:
            info(f'Creating directory: {options.GNN_dir}')
            os.makedirs(options.GNN_dir)

    if 'NTScore_dir' in io_options:
        if not options.NTScore_dir:
            error('Please provide a directory for the NTScore output.')
            optparser.print_help()
            sys.exit(1)
        if os.path.isdir(options.NTScore_dir):
            warning(f'The directory ({options.NTScore_dir}) you given already exists. It will be overwritten.')
        else:
            info(f'Creating directory: {options.NTScore_dir}')
            os.makedirs(options.NTScore_dir)


def write_io_options_memo(options: Values, io_options: Optional[List[str]]) -> None:
    """Write IO options to stdout.

    Ret: None
    """
    if io_options is None:
        return
    info('            -------- I/O options -------             ')
    if 'preprocessing_dir' in io_options:
        info(f'preprocessing output directory:  {options.preprocessing_dir}')
    if 'GNN_dir' in io_options:
        info(f'GNN output directory:  {options.GNN_dir}')
    if 'NTScore_dir' in io_options:
        info(f'NTScore output directory:  {options.NTScore_dir}')
    if 'dataset' in io_options:
        info(f'dataset: {options.dataset}')



from optparse import Values
from typing import Callable, Dict, List, Optional, Tuple, Type

import numpy as np
import pandas as pd
import torch
from numpy import ndarray
from scipy.sparse import load_npz
from torch_geometric.loader import DenseDataLoader

from ONTraC.data import SpatailOmicsDataset, create_torch_dataset
from ONTraC.log import *
from ONTraC.train import SubBatchTrainProtocol
from ONTraC.utils import get_rel_params, read_yaml_file
from ONTraC.utils.NTScore import (NTScore_table, get_niche_NTScore,
                                  niche_to_cell_NTScore)


def load_parameters(opt_validate_func: Callable, prepare_optparser_func: Callable) -> Values:
    """
    Load parameters
    :param opt_validate_func: validate function
    :param prepare_optparser_func: prepare optparser function
    :return: options
    """
    options = opt_validate_func(prepare_optparser_func())

    return options


def load_data(options: Values) -> Tuple[SpatailOmicsDataset, DenseDataLoader]:
    """
    Load data
    :param options: options
    :return: dataset, sample_loader
    """
    params = read_yaml_file(f'{options.preprocessing_dir}/samples.yaml')
    rel_params = get_rel_params(options, params)
    dataset = create_torch_dataset(options, rel_params)
    batch_size = options.batch_size if options.batch_size > 0 else len(dataset)
    sample_loader = DenseDataLoader(dataset, batch_size=batch_size)
    return dataset, sample_loader


def train(nn_model: Type[torch.nn.Module], options: Values, BatchTrain: Type[SubBatchTrainProtocol],
          device: torch.device, dataset: SpatailOmicsDataset, sample_loader: DenseDataLoader,
          inspect_funcs: Optional[List[Callable]], model_name: str) -> SubBatchTrainProtocol:
    info(message=f'{model_name} train start.')
    model = nn_model(input_feats=dataset.num_features,
                     hidden_feats=options.hidden_feats,
                     k=options.k,
                     exponent=options.beta)
    optimizer = torch.optim.Adam(model.parameters(), lr=options.lr)
    batch_train = BatchTrain(model=model, device=device, data_loader=sample_loader)  # type: ignore
    batch_train.save(path=f'{options.GNN_dir}/epoch_0.pt')

    loss_weight_args: dict[str, float] = {
        key: value
        for key, value in options.__dict__.items() if key.endswith('loss_weight')
    }

    batch_train.train(optimizer=optimizer,
                      inspect_funcs=inspect_funcs,
                      max_epochs=options.epochs,
                      max_patience=options.patience,
                      min_delta=options.min_delta,
                      min_epochs=options.min_epochs,
                      output=options.GNN_dir,
                      **loss_weight_args)
    batch_train.save(path=f'{options.GNN_dir}/model_state_dict.pt')
    return batch_train


def evaluate(batch_train: SubBatchTrainProtocol, model_name: str) -> None:
    """
    Evaluate process
    :return: None
    """
    info(message=f'{model_name} eval start.')
    loss_dict: Dict[str, np.floating] = batch_train.evaluate()  # type: ignore
    info(message=f'Evaluate loss, {repr(loss_dict)}')


def graph_pooling_output(ori_data_df: pd.DataFrame, dataset: SpatailOmicsDataset, rel_params: Dict,
                         consolidate_s_array: np.ndarray, output_dir: str) -> None:
    """
    Write the graph pooling results as the Niche cluster (max probability for each niche & cell).
    :param ori_data_df: pd.DataFrame, original data
    :param consolidate_s_array: np.ndarray, consolidate s array
    :return: None
    """

    consolidate_s_niche_df = pd.DataFrame()
    consolidate_s_cell_df = pd.DataFrame()
    for i, data in enumerate(dataset):
        # the slice of data in each sample
        slice_ = slice(i * data.x.shape[0], i * data.x.shape[0] + data.mask.sum())
        consolidate_s = consolidate_s_array[slice_]  # N x C
        consolidate_s_df_ = pd.DataFrame(consolidate_s,
                                         columns=[f'NicheCluster_{i}' for i in range(consolidate_s.shape[1])])
        consolidate_s_df_['Cell_ID'] = ori_data_df[ori_data_df['Sample'] == data.name]['Cell_ID'].values
        consolidate_s_niche_df = pd.concat([consolidate_s_niche_df, consolidate_s_df_], axis=0)

        # niche to cell matrix
        niche_weight_matrix = load_npz(rel_params['Data'][i]['NicheWeightMatrix'])
        niche_to_cell_matrix = (
            niche_weight_matrix /
            niche_weight_matrix.sum(axis=0)).T  # normalize by the all niches associated with each cell, N x N

        consolidate_s_cell = niche_to_cell_matrix @ consolidate_s
        consolidate_s_cell_df_ = pd.DataFrame(consolidate_s_cell,
                                              columns=[f'NicheCluster_{i}' for i in range(consolidate_s_cell.shape[1])])
        consolidate_s_cell_df_['Cell_ID'] = ori_data_df[ori_data_df['Sample'] == data.name]['Cell_ID'].values
        consolidate_s_cell_df = pd.concat([consolidate_s_cell_df, consolidate_s_cell_df_], axis=0)

    consolidate_s_niche_df = consolidate_s_niche_df.set_index('Cell_ID')
    consolidate_s_niche_df = consolidate_s_niche_df.loc[ori_data_df['Cell_ID'], :]
    consolidate_s_niche_df.to_csv(f'{output_dir}/niche_level_niche_cluster.csv.gz', index=True, index_label='Cell_ID', header=True)
    consolidate_s_niche_df['Niche_Cluster'] = consolidate_s_niche_df.values.argmax(axis=1)
    consolidate_s_niche_df['Niche_Cluster'].to_csv(f'{output_dir}/niche_level_max_niche_cluster.csv.gz',
                                                   index=True,
                                                   header=True)
    consolidate_s_cell_df = consolidate_s_cell_df.set_index('Cell_ID')
    consolidate_s_cell_df = consolidate_s_cell_df.loc[ori_data_df['Cell_ID'], :]
    consolidate_s_cell_df.to_csv(f'{output_dir}/cell_level_niche_cluster.csv.gz', index=True, index_label='Cell_ID', header=True)
    consolidate_s_cell_df['Niche_Cluster'] = consolidate_s_cell_df.values.argmax(axis=1)
    consolidate_s_cell_df['Niche_Cluster'].to_csv(f'{output_dir}/cell_level_max_niche_cluster.csv.gz',
                                                  index=True,
                                                  header=True)


def predict(output_dir: str, batch_train: SubBatchTrainProtocol, dataset: SpatailOmicsDataset,
            model_name: str) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    info(f'{model_name} predict start.')
    each_sample_loader = DenseDataLoader(dataset, batch_size=1)
    consolidate_flag = False
    consolidate_s_list = []
    consolidate_out = None
    consolidate_out_adj = None
    for data in each_sample_loader:  # type: ignore
        data = data.to(batch_train.device)  # type: ignore
        predict_result = batch_train.predict_dict(data=data)  # type: ignore
        for key, value in predict_result.items():
            np.savetxt(fname=f'{output_dir}/{data.name[0]}_{key}.csv.gz',
                       X=value.squeeze(0).detach().cpu().numpy(),
                       delimiter=',')

        # consolidate results
        if not consolidate_flag and ('s' in predict_result and 'out' in predict_result and 'out_adj' in predict_result):
            consolidate_flag = True
        if consolidate_flag:
            s = predict_result['s']
            out = predict_result['out']
            s = s.squeeze(0)
            consolidate_s_list.append(s)
            out_adj_ = torch.matmul(torch.matmul(s.T, data.adj.squeeze(0)), s)
            consolidate_out_adj = out_adj_ if consolidate_out_adj is None else consolidate_out_adj + out_adj_
            consolidate_out = out.squeeze(
                0) * data.mask.sum() if consolidate_out is None else consolidate_out + out.squeeze(0) * data.mask.sum()

    if consolidate_flag:
        # consolidate out
        nodes_num = 0
        for data in each_sample_loader:  # type: ignore
            nodes_num += data.mask.sum()
        consolidate_out = consolidate_out / nodes_num  # type: ignore
        consolidate_out_array = consolidate_out.detach().cpu().numpy()
        np.savetxt(fname=f'{output_dir}/consolidate_out.csv.gz', X=consolidate_out_array, delimiter=',')
        # consolidate s
        consolidate_s = torch.cat(consolidate_s_list, dim=0)
        # consolidate out_adj
        ind = torch.arange(consolidate_s.shape[-1], device=consolidate_out_adj.device)
        consolidate_out_adj[ind, ind] = 0
        d = torch.einsum('ij->i', consolidate_out_adj)
        d = torch.sqrt(d)[:, None] + 1e-15
        consolidate_out_adj = (consolidate_out_adj / d) / d.transpose(0, 1)
        consolidate_s_array = consolidate_s.detach().cpu().numpy()
        consolidate_out_adj_array = consolidate_out_adj.detach().cpu().numpy()
        np.savetxt(fname=f'{output_dir}/consolidate_s.csv.gz', X=consolidate_s_array, delimiter=',')
        np.savetxt(fname=f'{output_dir}/consolidate_out_adj.csv.gz', X=consolidate_out_adj_array, delimiter=',')

        return consolidate_s_array, consolidate_out_adj_array
    else:
        return None, None


def NTScore(options: Values, dataset: SpatailOmicsDataset, consolidate_s_array: ndarray,
            consolidate_out_adj_array: ndarray) -> None:
    """
    Pseudotime calculateion process
    :param options: options
    :param consolidate_s_array: consolidate s array
    :param consolidate_out_adj_array: consolidate out adj array
    :return: None
    """

    params = read_yaml_file(f'{options.preprocessing_dir}/samples.yaml')
    rel_params = get_rel_params(options, params)

    niche_cluster_score, niche_level_NTScore = get_niche_NTScore(niche_cluster_loading=consolidate_s_array,
                                                                 niche_adj_matrix=consolidate_out_adj_array)
    cell_level_NTScore, all_niche_level_NTScore_dict, all_cell_level_NTScore_dict = niche_to_cell_NTScore(
        dataset=dataset, rel_params=rel_params, niche_level_NTScore=niche_level_NTScore)

    np.savetxt(fname=f'{options.NTScore_dir}/niche_cluster_score.csv.gz', X=niche_cluster_score, delimiter=',')
    np.savetxt(fname=f'{options.NTScore_dir}/niche_NTScore.csv.gz', X=niche_level_NTScore, delimiter=',')
    np.savetxt(fname=f'{options.NTScore_dir}/cell_NTScore.csv.gz', X=cell_level_NTScore, delimiter=',')

    NTScore_table(options=options,
                  rel_params=rel_params,
                  all_niche_level_NTScore_dict=all_niche_level_NTScore_dict,
                  all_cell_level_NTScore_dict=all_cell_level_NTScore_dict)



import sys
import time


def get_current_time() -> str:
    return time.strftime('%H:%M:%S', time.localtime())


def write_direct_message(message: str):
    curr_time_str = get_current_time()
    sys.stdout.write(f'{curr_time_str} --- {message}\n')
    sys.stdout.flush()


def debug(message: str):
    write_direct_message(f'DEBUG: {message}')


def info(message: str):
    write_direct_message(f'INFO: {message}')


def write_direct_message_err(message: str):
    curr_time_str = get_current_time()
    sys.stderr.write(f'{curr_time_str} --- {message}\n')
    sys.stderr.flush()


def warning(message: str):
    write_direct_message_err(f'WARNING: {message}')


def error(message: str):
    write_direct_message_err(f'ERROR: {message}')


def critical(message: str):
    write_direct_message_err(f'CRITICAL: {message}')


__all__ = ['debug', 'info', 'warning', 'error', 'critical']
